/*----------------------------------------------------------------------*/ 
/*  Name           : mlp.c                                              */
/*  Version        : 1.0                                                */
/*  Creation       : 13/05/01                                           */
/*  Last update    : 05/06/24                                           */ 
/*  Subject        : Implementation of an MLP                           */
/*  Author         : Yann Guermeur                                      */
/*----------------------------------------------------------------------*/


#include <stdio.h>
#include <string.h>
#include <math.h>
#include <ctype.h>
#include <stdlib.h>
#include <time.h>
#include "biblio.h"
#include "algebre.h"
#include <sys/types.h>
#include <unistd.h>

#define true 1
#define false 0
#define inf(a,b) ((a)<=(b)?(a):(b))
#define sup(a,b) ((a)>=(b)?(a):(b))

#define taille 81
#define pas 100

/* Global variables */

FILE *f1;

char fichier_fichcom[taille], fichier_data[taille], conf[taille],
 fichier_matrix_init[taille], fichier_matrix_fin[taille],
 fichier_output[taille];

unsigned short seed[3];

int status;

long i, j, k, nb_data, hls=4, ops = 3, Q = 3, indice_exemple = 1, random_num,
 learning=false, iter, nbiter, exemple, dim_input, **y, choice[3], Q_check;

double **dcdxl, epsilon=0.01, **des, **dcdxk, **uweight, bound=0.0,
 **x_i, **fbias, **ubias, **hidden, epsilon_min=0.001,
 **fweight, **output, **mat, **X, cost=0.0;

/* Functions included in this program */

void caract_db();
void read_data();
void weights_init();
void compute_xi();
void compute_des();
void compute_hidden_outputs();
void compute_top_outputs();
void compute_dcdxl();
void compute_dcdxk();
void modif_uweight();
void modif_fweight();
void modif_bias();
void write_matrixes();
void compute_cost();
void eval_rec();
void network_use();
void learning_proc();
void Pause(char *message);

int main (int argc, char *argv[])

/* MLP - objective function: cross entropy */

{

status = system("clear");
srand48(getpid());
strcpy(fichier_fichcom, argv[1]);

printf("\n      Name of the file of parameters: %s", fichier_fichcom);

caract_db();
read_data();

weights_init();

if(learning==true)
  {
  for(iter=1; iter<=nbiter; iter++)
    {
    learning_proc(); 
    if(iter%pas == 0)
      {
      printf("Iteration %5ld\n", iter);
      write_matrixes();
      }
    }
  }
else
  {
  network_use();
  display_stats(fichier_data, nb_data, Q, mat, cost);
  }

}

void Pause(char *message)

{

char c;

static int n = 0;

if(message!=NULL)
  printf("%s\n",message);

printf("\a\n=> New Line to proceed...\n");

n = read(0, &c, 1);

while(c != '\n')
  n = read(0, &c, 1);

return;

}

void caract_db()

/* Characterization of the problem */

{

if((f1=fopen(fichier_fichcom, "r"))==NULL)
  exit(0);

choice[1]=getc(f1);

switch(choice[1])
  {
  case 'o': learning=true; break;
  case 'n': break;
  default: printf("\n\nCoding problem...\n\n"); exit(0);
  }

choice[2]=getc(f1);

status = fscanf(f1, "%ld", &Q);
status = fscanf(f1, "%s", fichier_data);
status = fscanf(f1, "%s", fichier_matrix_init);
status = fscanf(f1, "%s", fichier_matrix_fin);
status = fscanf(f1,"%ld", &hls);
status = fscanf(f1,"%lf", &epsilon);
status = fscanf(f1,"%lf", &bound);
status = fscanf(f1,"%ld" ,&nbiter);
status = fscanf(f1,"%s" ,fichier_output);

printf("\n                Name of the data set: %s", fichier_data);
if(learning == true)
  {
  epsilon_min = epsilon * 0.1;
  printf("\n Name of the file of initial weights: %s", fichier_matrix_init);
  printf("\n   Name of the file of final weights: %s", fichier_matrix_fin);
  printf("\n  Minimal value of the gradient step: %lf\n", epsilon_min);
  }
else
  {
  printf("\n         Name of the file of weights: %s", fichier_matrix_init);
  printf("\nName of the file of computed outputs: %s", fichier_output);
  }

fclose(f1);

ops=Q;

mat = matrix(Q+1, Q+1);

for(i=1; i<=Q; i++)
  for(j=1; j<=Q; j++)
    mat[i][j] = 0.0;

hidden = matrix(hls+1, 1);
dcdxk = matrix(hls+1, 1);

for(i=1; i<=hls; i++)
  {
  hidden[i][1] = 0.0;
  dcdxk[i][1] = 0.0;
  }

output = matrix(ops+1, 1);
des = matrix(ops+1, 1);
dcdxl = matrix(ops+1, 1);

for(i=1; i<=ops; i++)
  {
  output[i][1] = 0.0;
  des[i][1] = 0.0;
  dcdxl[i][1] = 0.0;
  }

}

void read_data()

/* Reading the data set */

{

long y_min = Q, y_max = 0;

if((f1=fopen(fichier_data, "r"))==NULL)
  {
  printf("\nData file: %s not found...\n", fichier_data);
  exit(0);
  }

status = fscanf(f1, "%ld", &nb_data);
status = fscanf(f1, "%ld", &dim_input);
status = fscanf(f1, "%ld", &Q_check);

if(Q_check != Q)
  {
  printf("\n\nIncoherence in the numbers of categories...");
  exit(0);
  }


X = matrix(nb_data+1, dim_input+1);
x_i = matrix(dim_input+1, 1);
y = matrix_l(nb_data+1, 1);

for(i=1; i<=nb_data; i++)
  {
  for(j=1; j<=dim_input; j++)
    status = fscanf(f1, "%lf", &X[i][j]);

  status = fscanf(f1, "%ld", &y[i][1]);

  if(y[i][1] < y_min)
     y_min = y[i][1];

  if(y[i][1] > y_max)
     y_max = y[i][1];
  }

fclose(f1);

/*
printf("\nIndices of categories ranging from %ld to %ld\n", y_min, y_max);
Pause("");
*/

if(learning==true)
  if(y_min != 1 || y_max != Q)
    {
    printf("\nWrong numbering of the categories\n");
    exit(0);
    }
else
  if(y_min < 1 || y_max > Q)
    {
    printf("\nWrong numbering of the categories\n");
    exit(0);
    }

}

void weights_init()

/* Initialization of the weights */

{

fweight = matrix(hls+1, dim_input+1);
uweight  = matrix(ops+1, hls+1);

fbias = matrix(hls+1, 1);
ubias  = matrix(ops+1, 1);

if(choice[2] == 'o')
  {
  if((f1=fopen(fichier_matrix_init, "r"))==NULL)
    {
    printf("\nFile of the weights, %s: cannot be found...\n", 
    fichier_matrix_init);
    exit(0);
    }
  else
    printf("\n\nReading the values of the weights...\n");

  for(i=1; i<=hls; i++)
    for(j=1; j<=dim_input; j++)
      status = fscanf(f1, "%lf", &fweight[i][j]);

  for(i=1; i<=ops; i++)
    for(j=1; j<=hls; j++)
      status = fscanf(f1, "%lf", &uweight[i][j]);

  for(i=1; i<=hls; i++)
    status = fscanf(f1, "%lf", &fbias[i][1]);

  for(i=1; i<=ops; i++)
    status = fscanf(f1, "%lf", &ubias[i][1]);

  fclose(f1);
  } 
else
  {

  printf("\nRandom initialization of the weights...\n\n");

  for(i=1; i<=hls; i++)
    for(j=1; j<=dim_input; j++)
      fweight[i][j] = bound*(drand48()-0.5);

  for(i=1; i<=ops; i++)
    for(j=1; j<=hls; j++)
      uweight[i][j] = bound*(drand48()-0.5);

  for(i=1; i<=hls; i++)
    fbias[i][1] = (bound*(drand48()-0.5))/10.0;
 
  for(i=1; i<=ops; i++)
    ubias[i][1] = (bound*(drand48()-0.5))/10.0;
  }

}

void compute_xi()

/* Computation of the input vector */

{

if(learning == true)
  {
  random_num = nrand48(seed);
  indice_exemple = (random_num % nb_data) + 1;
  }
else
  indice_exemple = exemple;

if((indice_exemple < 1) || (indice_exemple > nb_data))
  {
  printf("\n\nWrong index of example...\n\n");
  exit(0);
  }

for(k=1; k<=dim_input; k++)
  x_i[k][1] = X[indice_exemple][k];

}

void compute_des()

/* Computation of desired output */

{

for(k=1; k<=Q; k++)
  des[k][1] = (y[indice_exemple][1] == k) ? 1.0 : 0.0;

}

void compute_hidden_outputs()

/* Computation of the output of the hidden layer */

{

mult_mat(fweight, x_i, hidden, hls, dim_input, 1);
add_mat1(hidden, fbias, hls, 1);

for(i=1; i<=hls; i++)
  hidden[i][1] = tanh(hidden[i][1]);

}

void compute_top_outputs()

/* Computation of the output of the network */

{

double sum = 0.0;

mult_mat(uweight, hidden, output, ops, hls, 1);
add_mat1(output, ubias, ops, 1);

for(i=1; i<=ops; i++)
  {
  output[i][1] = exp(output[i][1]);
  sum += output[i][1];
  }

for(i=1; i<=ops; i++)
  output[i][1] /= sum;

}

void compute_dcdxl()

/* Computation of the partial derivatives dc/dxl */

{

sub_mat(output, des, dcdxl, ops, 1);

}

void compute_dcdxk()

/* Computation of the partial derivatives dc/dxk */

{

for(i=1; i<=hls; i++)
  dcdxk[i][1] = 0.0;

for(i=1; i<=ops; i++)
  for(j=1; j<=hls; j++)
    dcdxk[j][1] += dcdxl[i][1] * uweight[i][j];

for(i=1; i<=hls; i++)
  dcdxk[i][1] *= tanh_prim(hidden[i][1]);

}

void modif_uweight()

/* Update of the weights of the top layer */

{

for(i=1; i<=ops; i++)
  for(j=1; j<=hls; j++)
    uweight[i][j] -= epsilon * dcdxl[i][1] * hidden[j][1];

}
 
void modif_fweight()

/* Update of the weights of the bottom layer */

{

for(i=1; i<=hls; i++)
  for(j=1; j<=dim_input; j++)
    fweight[i][j] -= epsilon * dcdxk[i][1] * x_i[j][1];

}

void modif_bias()

/* Update of the weights of the bias cell */

{

for(i=1; i<=hls; i++)
  fbias[i][1] -= epsilon * dcdxk[i][1];

for(i=1; i<=ops; i++)
  ubias[i][1] -= epsilon * dcdxl[i][1];

}

void compute_cost()

/* Computation of the objective function (cross entropy) */

{

if(output[y[exemple][1]][1] > 0.0)
  cost -= log(output[y[exemple][1]][1]);
else
  {
  printf("\nExample: %ld, objective function is infinite...\n\n",
         exemple);
  exit(0);
  }

}

void eval_rec()

/* Classification of a pattern */

{

long class_pred = 0;

double maximum = 0.0;

for(k=1; k<=Q; k++)
  if(output[k][1] > maximum)
    {
    class_pred = k;
    maximum = output[k][1];
    }

for(k=1; k<=Q; k++)
  fprintf(f1, "%lf%c", output[k][1], (k==Q) ? '\n':' ');

mat[y[exemple][1]][class_pred]++;

}

void learning_proc()

/* Implementation of an epoch */

{

if(epsilon >= epsilon_min)
  epsilon *= 0.998;

for(exemple=1; exemple<=nb_data; exemple++)
  {
  compute_xi();
  compute_des();
  compute_hidden_outputs();
  compute_top_outputs();
  compute_dcdxl();
  compute_dcdxk();
  modif_uweight();
  modif_fweight();
  modif_bias();
  }

}

void network_use()

/* Assessment of the network on a test set */

{

if((f1=fopen(fichier_output, "w"))==NULL)
  exit(0);

for(exemple=1; exemple<=nb_data; exemple++)
  {
  compute_xi();
  compute_des();
  compute_hidden_outputs();
  compute_top_outputs();
  compute_cost();
  eval_rec();
  }

fclose(f1);

}

void write_matrixes()

/* Save of network weights */

{

if((f1=fopen(fichier_matrix_fin,"w"))==NULL)
  exit(0);

for(i=1; i<=hls; i++)
  for(j=1; j<=dim_input; j++)
    fprintf(f1,"%lf\n", fweight[i][j]);

for(i=1; i<=ops; i++)
  for(j=1; j<=hls; j++)
    fprintf(f1,"%lf\n", uweight[i][j]);

for(i=1; i<=hls; i++)
  fprintf(f1,"%lf\n", fbias[i][1]);

for(i=1; i<=ops; i++)
  fprintf(f1,"%lf\n", ubias[i][1]);

fclose(f1);

}
